import { ExpressionVisitor, NodeType, TypeGuard, } from "@surface/expression";
export default class ObserverVisitor extends ExpressionVisitor {
    paths = [];
    brokenPath = false;
    stack = [];
    static observe(expression) {
        const visitor = new ObserverVisitor();
        visitor.visit(expression);
        visitor.commit();
        return visitor.paths;
    }
    commit() {
        if (this.stack.length > 1) {
            this.paths.push([...this.stack]);
        }
        this.stack = [];
    }
    reset() {
        this.brokenPath = false;
        this.stack = [];
    }
    rollback() {
        this.brokenPath = true;
        this.stack = [];
    }
    visitArrayPattern(expression) {
        for (const element of expression.elements) {
            if (element && !TypeGuard.isIdentifier(element)) {
                this.visit(element);
            }
        }
        return expression;
    }
    visitArrowFunctionExpression(expression) {
        for (const parameter of expression.parameters) {
            if (!TypeGuard.isIdentifier(parameter)) {
                this.visit(parameter);
            }
        }
        return expression;
    }
    visitAssignmentExpression(expression) {
        this.visit(expression.right);
        return expression;
    }
    visitAssignmentPattern(expression) {
        this.visit(expression.right);
        return expression;
    }
    visitAssignmentProperty(expression) {
        if (TypeGuard.isAssignmentPattern(expression.value)) {
            this.visitAssignmentPattern(expression.value);
        }
        return expression;
    }
    visitCallExpression(expression) {
        this.rollback();
        this.visit(expression.callee);
        this.reset();
        expression.arguments.forEach(this.visit.bind(this));
        return expression;
    }
    visitLogicalExpression(expression) {
        if (expression.operator == "??") {
            this.visit(expression.left);
        }
        else {
            super.visitLogicalExpression(expression);
        }
        return expression;
    }
    visitConditionalExpression(expression) {
        this.visit(expression.test);
        return expression;
    }
    visitIdentifier(expression) {
        if (expression.name != "undefined") {
            this.stack.unshift(expression.name);
            this.commit();
        }
        return expression;
    }
    visitMemberExpression(expression) {
        if (expression.computed && expression.property.type != NodeType.Literal) {
            this.reset();
            this.visit(expression.property);
        }
        else if (expression.optional) {
            this.rollback();
        }
        else if (!this.brokenPath) {
            const key = TypeGuard.isIdentifier(expression.property) ? expression.property.name : `${expression.property.evaluate({})}`;
            this.stack.unshift(key);
        }
        this.brokenPath = expression.optional;
        this.visit(expression.object);
        return expression;
    }
    visitNewExpression(expression) {
        this.rollback();
        this.visit(expression.callee);
        this.reset();
        expression.arguments.forEach(this.visit.bind(this));
        return expression;
    }
    visitParenthesizedExpression(expression) {
        this.reset();
        this.visit(expression.argument);
        return expression;
    }
    visitProperty(expression) {
        if (expression.computed) {
            this.visit(expression.key);
        }
        this.visit(expression.value);
        return expression;
    }
    visitRestElement(expression) {
        if (!TypeGuard.isIdentifier(expression.argument)) {
            this.visit(expression.argument);
        }
        return expression;
    }
    visitTaggedTemplateExpression(expression) {
        this.rollback();
        this.visit(expression.callee);
        this.reset();
        this.visit(expression.quasi);
        return expression;
    }
    visitThisExpression(expression) {
        this.stack.unshift("this");
        this.commit();
        return expression;
    }
    visitUpdateExpression(expression) {
        this.rollback();
        this.visit(expression.argument);
        return expression;
    }
}
